/* iRiver iFP 7XX Emulator
   Copyright (C) 2006 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>

#include "flash.h"
#include "log.h"

static int flash_read_page_oob(struct flash *flash, int page) {
  flash->only_oob = 1;
  flash->page = page;
  if (fseek(flash->fp, flash->offset + page * 64, SEEK_SET) == -1)
    return -1;
  if (fread(flash->reg + 2048, 1, 64, flash->fp) != 64) return -1;
  return 0;
}

static inline int flash_n_pages(struct flash *flash) {
  switch (flash->model) {
  case FLASH_256_8:
    return 128 * 1024;
  case FLASH_512_8:
    return 256 * 1024;
  case FLASH_1024_8:
    return 512 * 1024;
  default:
    return 0;
  }
}

static inline int flash_row_mask(struct flash *flash) {
  switch (flash->model) {
  case FLASH_256_8:
    return 1;
  case FLASH_512_8:
    return 3;
  case FLASH_1024_8:
    return 7;
  default:
    return 0;
  }
}

static inline int flash_id(struct flash *flash) {
  switch (flash->model) {
  case FLASH_256_8:
    return 0xda;
  case FLASH_512_8:
    return 0xdc;
  case FLASH_1024_8:
    return 0xdf;
  }
}

static int flash_read_page_rest(struct flash *flash) {
  if (flash->only_oob) {
    if (fseek(flash->fp,
	      flash->offset + flash->page * 2048 + flash_n_pages(flash) * 64,
	      SEEK_SET) == -1)
      return -1;
    if (fread(flash->reg, 1, 2048, flash->fp) != 2048) return -1;
  }
  flash->only_oob = 0;
}

static int flash_read_page_byte(struct flash *flash, int ptr) {
  if (ptr < 2048)
    flash_read_page_rest(flash);
  return flash->reg[ptr];
}

static int flash_write_page(struct flash *flash, int page) {
  if (fseek(flash->fp, flash->offset + page * 64, SEEK_SET) == -1)
    return -1;
  if (fwrite(flash->reg + 2048, 1, 64, flash->fp) != 64) return -1;
  if (fseek(flash->fp,
	    flash->offset + page * 2048 + flash_n_pages(flash) * 64,
	    SEEK_SET) == -1)
    return -1;
  if (fwrite(flash->reg, 1, 2048, flash->fp) != 2048) return -1;
  return 0;
}

uint8 flash_read(struct flash *flash, int adr) {
  uint8 r;
  switch (adr & 0xc) {
  case 0:
    if (flash->ptr >= 2112)
      flash->ptr = 0;
    r = flash_read_page_byte(flash, flash->ptr);
    flash->ptr++;
    logmsg(LOG_FLASH, 20, "flash: read %x from %x\n", r, adr);
    return r;
  default:
    logmsg(LOG_FLASH, 5, "flash: read from %x\n", adr);
  }
  return ~0;
}

static void flash_read_cmd(struct flash *flash) {
  int i;
  int col = flash->adr[0] | ((flash->adr[1] & 0xf) << 8);
  int row =  flash->adr[2] | (flash->adr[3] << 8)
    | ((flash->adr[4] & flash_row_mask(flash)) << 16);

  logmsg(LOG_FLASH, 5, "flash: read, address %6x/%3x\n", row, col);

  flash_read_page_oob(flash, row);

  flash->ptr = col;
}

static void flash_page_program(struct flash *flash) {
  int i;
  int row =  flash->adr[2] | (flash->adr[3] << 8)
    | ((flash->adr[4] & flash_row_mask(flash)) << 16);

  logmsg(LOG_FLASH, 5, "flash: program, address %6x\n", row);

  flash_read_page_rest(flash);
  flash_write_page(flash, row);
}

static void flash_block_erase(struct flash *flash) {
  int row =  flash->adr[0] | (flash->adr[1] << 8)
    | ((flash->adr[2] & flash_row_mask(flash)) << 16);
  int i;

  logmsg(LOG_FLASH, 5, "flash: op block erase, address %6x\n", row);
  row &= ~0x3f;

  for (i = 0; i < 2112; i++)
    flash->reg[i] = 0xff;

  for (i = 0; i < 64; i++)
    flash_write_page(flash, row + i);
}

void flash_write(struct flash *flash, int adr, uint8 val) {
  int i;
  switch (adr & 0xc) {
  case 0:
    flash_read_page_rest(flash);
    if (flash->ptr >= 2112)
      flash->ptr = 0;
    flash->reg[flash->ptr] = val;
    flash->ptr++;
    logmsg(LOG_FLASH, 20, "flash: write %x to %x\n", val, adr);
    break;
  case 4:
    flash->adr_ptr = 0;
    logmsg(LOG_FLASH, 19, "flash: command %x\n", val);

    switch (val) {
    case 0x90:
      flash->cmd = val;
      logmsg(LOG_FLASH, 9, "flash: read id\n");
      flash->reg[0] = 0xec;
      flash->reg[1] = flash_id(flash);
      flash->reg[2] = 0xff;
      flash->reg[3] = 0x15;
      flash->ptr = 0;
      flash->only_oob = 0;
      break;
    case 0x70:
      flash->cmd = val;
      flash->reg[0] = 0xe0;
      flash->ptr = 0;
      flash->only_oob = 0;
      break;
    case 0x00:
    case 0x05:
    case 0x60:
    case 0x80:
    case 0x85:
      flash->cmd = val;
      break;
    case 0x10:
      flash_page_program(flash);
      break;
    case 0x30:
    case 0x35:
      flash_read_cmd(flash);
      break;
    case 0xd0:
      flash_block_erase(flash);
      break;
    case 0xe0:
      i = flash->adr[0] | ((flash->adr[1] & 0xf) << 8);
      flash->ptr = i;
      logmsg(LOG_FLASH, 10, "flash: read random, column %3x\n", i);
      break;
    default:
      break;
    }
    break;
  case 8:
    logmsg(LOG_FLASH, 19, "flash: address %x\n", val);
    if (flash->adr_ptr >= 5)
      flash->adr_ptr = 0;

    flash->adr[flash->adr_ptr] = val;
    flash->adr_ptr++;

    if (flash->cmd == 0x80 || flash->cmd == 0x85) {
      flash->ptr = flash->adr[0] | ((flash->adr[1] & 0xf) << 8);
    }

    if (flash->cmd == 0x80 && flash->adr_ptr == 5) {
      int row = flash->adr[2] | (flash->adr[3] << 8)
	| ((flash->adr[4] & flash_row_mask(flash)) << 16);
      flash_read_page_oob(flash, row);
    }

    break;
  default:
    logmsg(LOG_FLASH, 5, "flash: write %x to %x\n", val, adr);
  }
}

struct flash *create_flash(int model, char *fname, int offset) {
  struct flash *flash;

  flash = xmalloc(sizeof(struct flash));
  if (!flash)
    return NULL;

  flash->ptr = 0;
  flash->only_oob = 0;
  flash->model = model;
  flash->fp = fopen(fname, "r+b");
  if (!flash->fp) {
    error_sys("%s: ", fname);
    free(flash);
    return NULL;
  }
  flash->busy = 0;
  flash->offset = offset;
  return flash;
}

void free_flash(struct flash *flash) {
  if (flash) {
    if (flash->fp)
      fclose(flash->fp);
    free(flash);
  }
}

