/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <stdio.h>
#include <stdlib.h>

#include "memory.h"
#include "cpu.h"
#include "hardware.h"
#include "log.h"
#include "lcd.h"
#include "usb.h"
#include "flash.h"

unsigned int tick;

int tick_before(unsigned int a, unsigned int b) {
  return (unsigned int)(b - a) < (1 << 30);
}

static uint32 gpio_read(struct hardware *hardware, uint32 adr) {
  int n = (adr >> 6) & 7;
  int r = adr & 0x3f;
  int i;
  uint32 res;

  switch (r) {
  case 0:
    logmsg(LOG_GPIO, 20, "gpio(%d): read from register %x\n", n, r);
    res = 0;
    for (i = 31; i >= 0; i--) {
      struct gpio_handler *h = &(hardware->gpio_handler[n][i]);
      if (h->get) {
	res = (res << 1) | (*(h->get))(h->data, n, i);
      } else
	res = (res << 1) | 1;
    }
    return res;
  default:
    logmsg(LOG_GPIO, 5, "gpio(%d): read from register %x\n", n, r);
    return ~0;
  }
}

static void gpio_notify_handlers(struct hardware *hardware,
				 int port, uint32 old, uint32 new) {
  int i;
  for (i = 0; i < 32; i++)
    if ((old ^ new) & (1 << i)) {
      struct gpio_handler *h = &(hardware->gpio_handler[port][i]);
      if (h->set)
	(*h->set)(h->data, port, i, (new >> i) & 1);
    }
      
}

int gpio_get_bit(struct hardware *hardware, int port, int bit) {
  return (hardware->gpio_out[port] >> bit) & 1;
}

void gpio_install_handler(struct hardware *hardware, int port, int bit,
			  void *data,
			  int (*get)(void *, int, int),
			  void (*set)(void *, int, int, int)) {
  hardware->gpio_handler[port][bit].data = data;
  hardware->gpio_handler[port][bit].get = get;
  hardware->gpio_handler[port][bit].set = set;
}

static void gpio_direction_changed(struct hardware *hardware, int port,
				   uint32 mask, uint32 new) {
  int i;
  for (i = 0; i < 32; i++)
    if (mask & (1 << i))
      logmsg(LOG_GPIO, 10, "gpio: bit %d/%d set to %s\n",
	     port, i, (new & (1 << i)) ? "output" : "input");
}

static void gpio_write(struct hardware *hardware, uint32 adr, uint32 val) {
  int n = (adr >> 6) & 7;
  int r = adr & 0x3c;
  uint32 old;

  switch (r) {
  case 0x14:
    old = hardware->gpio_out[n];
    hardware->gpio_out[n] |= val;
    gpio_notify_handlers(hardware,
			 n, old, hardware->gpio_out[n]);
    break;
  case 0x18:
    old = hardware->gpio_out[n];
    hardware->gpio_out[n] &= ~val;
    gpio_notify_handlers(hardware,
			 n, old, hardware->gpio_out[n]);
    break;
  case 0x24:
    hardware->gpio_dir[n] |= val;
    gpio_direction_changed(hardware, n,
			   val, hardware->gpio_dir[n]);
    break;
  case 0x28:
    hardware->gpio_dir[n] &= ~val;
    gpio_direction_changed(hardware, n,
			   val, hardware->gpio_dir[n]);
    break;
  default:
    logmsg(LOG_GPIO, 5, "gpio(%d): write %x to register %x\n", n, val, r);
  }
}

static int flash_ce[4] = {0x20, 0x02, 0x10, 0x08};

static int selected_flash(struct hardware *hardware) {
  int i, r;
  uint32 p5 = hardware->gpio_out[5];
  r = -1;
  for (i = 0; i < 4; i++)
    if (!(p5 & flash_ce[i])) {
      if (r > 0)
	logmsg(LOG_FLASH, 5, "flash: more than one chip selected\n");
      else
	r = i;
    }
  return r;
}

static int flash_get_busy(void *data, int port, int bit) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;

  chip = selected_flash(hardware);
  if (chip == -1)
    return 1;

  flash = hardware->flash[chip];
  if (!flash)
    return 1;

  if (!flash->busy)
    return 1;

  if (tick_before(flash->ready_tick, tick)) {
    flash->busy = 0;
    return 1;
  }

  return 0;
}

static uint32 prog_flash_read_half_word(void *data, uint32 adr) {
  struct prog_flash *flash = data;
  logmsg(LOG_FW_FLASH, 10, "prog_flash: read from %x\n", adr);
  adr = adr & 0xffffe;

  switch (flash->state) {
  case 0:
    return flash->content[adr] | (flash->content[adr + 1] << 8);

  case 0x90:
    switch (adr) {
    case 0:
      return 0xbf;
    case 2:
      return 0x2781;
    default:
      return 0xffff;
    }
  default:
    return 0xffff;
  }
}

static uint8 prog_flash_read_byte(void *data, uint32 adr) {
  uint8 res;
  if (adr & 1)
    res = prog_flash_read_half_word(data, adr & ~1) >> 8;
  else
    res = prog_flash_read_half_word(data, adr) & 0xff;
  logmsg(LOG_FW_FLASH, 10, "prog_flash: read byte %x\n", res);
  return res;
}

static uint32 prog_flash_read_word(void *data, uint32 adr) {
  return prog_flash_read_half_word(data, adr & ~3)
    | (prog_flash_read_half_word(data, (adr & ~3) + 2) << 16);
}

static int prog_flash_update(struct prog_flash *flash, int start, int size) {
  if (fseek(flash->fp, start, SEEK_SET))
    return -1;
  fwrite(flash->content + start, 1, size, flash->fp);
  return 0;
}

static void prog_flash_sector_erase(struct prog_flash *flash, int sector) {
  sector = sector & 0xff;
  logmsg(LOG_FW_FLASH, 5, "prog_flash: sector %x erase\n", sector);
  memset(flash->content + sector * 4096, 0xff, 4096);
  prog_flash_update(flash, sector * 4096, 4096);
}

static void prog_flash_block_erase(struct prog_flash *flash, int block) {
  block = block & 0xf;
  logmsg(LOG_FW_FLASH, 5, "prog_flash: block %x erase\n", block);
  memset(flash->content + block * 64 * 1024, 0xff, 64 * 1024);
  prog_flash_update(flash, block * 64 * 1024, 64 * 1024);
}

static void prog_flash_chip_erase(struct prog_flash *flash) {
  logmsg(LOG_FW_FLASH, 5, "prog_flash: chip erase\n");
  memset(flash->content, 0xff, 1024 * 1024);
  prog_flash_update(flash, 0, 1024 * 1024);
}

static void prog_flash_write_half_word(void *data, uint32 adr, uint32 val) {
  struct prog_flash *flash = data;
  int adr15 = (adr & 0xffff) >> 1;
  int val8 = val & 0xff;

  logmsg(LOG_FW_FLASH, 10, "prog_flash: write %x to %x\n", val, adr);

  switch (flash->state) {
  case 0:
    if (adr15 == 0x5555 && val8 == 0xaa)
      flash->state = 1;
    break;

  case 1:
    if (adr15 == 0x2aaa && val8 == 0x55)
      flash->state = 2;
    else
      flash->state = 0;
    break;

  case 2:
    if (adr15 == 0x5555) {
      switch (val8) {
      case 0xa0:
      case 0x80:
      case 0x90:
      case 0x98:
	flash->state = val & 0xff;
	break;
      default:
	flash->state = 0;
	break;
      }
    } else
      flash->state = 0;
    break;

  case 0x80:
    /* erase mode */
    if (adr15 == 0x5555 && val8 == 0xaa)
      flash->state = 0x81;
    else
      flash->state = 0;
    break;

  case 0x81:
    if (adr15 == 0x2aaa && val8 == 0x55)
      flash->state = 0x82;
    else
      flash->state = 0;
    break;

  case 0x82:
    switch (val8) {
    case 0x30:
      prog_flash_sector_erase(flash, adr >> 12);
      flash->state = 0;
      break;
    case 0x50:
      prog_flash_block_erase(flash, adr >> 16);
      flash->state = 0;
      break;
    case 0x10:
      if (adr15 == 0x5555)
	prog_flash_chip_erase(flash);
      flash->state = 0;
      break;
    default:
      flash->state = 0;
      break;
    }
    break;
  case 0x90:
    /* software id mode */
    if (adr15 == 0x5555 && val8 == 0xf0)
      flash->state = 0;
    break;

  case 0x98:
    /* TODO: CFI query */
    flash->state = 0;
    break;

  case 0xa0:
    adr = adr & 0xffffe;
    flash->content[adr] = val & 0xff;
    flash->content[adr + 1] = val >> 8;
    prog_flash_update(flash, adr, 2);
    flash->state = 0;
    break;

  default:
    flash->state = 0;
    break;
  }
}

static uint32 read_803(struct hardware *hardware, uint32 adr) {
  logmsg(LOG_MEMORY, 5, "irq: read from %x\n", adr);

  if ((adr & 0xfff) < 0x500)
    return hardware->irq_reg[(adr & 0xfff) >> 2];
  else
    return ~0;
}

static void write_803(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_MEMORY, 5, "irq: write %x to %x\n", val, adr);

  if ((adr & 0xfff) < 0x500)
    hardware->irq_reg[(adr & 0xfff) >> 2] = val;
}

int check_irq(void *data) {
  struct hardware *hardware = data;
  int i;
  
  if (hardware->irq_request) {
    for (i = 0; i < 32; i++)
      if (hardware->irq_request & (1 << i)) {
	hardware->irq_request &= ~(1 << i);
	hardware->irq_reg[0x40] = i * 8;
	return 1;
      }
  }
  return 0;
}

void set_irq_request(struct hardware *hardware, int i) {
  if (hardware->irq_reg[0x100 + i] & 0x10000)
    hardware->irq_request |= (1 << i);
}

static void timer_irq(struct event *e) {
  struct hardware *hardware = e->data;
  set_irq_request(hardware, 5);
  if (hardware->timer_period) {
    e->tick = tick + hardware->timer_period;
    add_event(e);
  } else
    hardware->timer_active = 0;
}

static void timer_write(struct hardware *hardware,
			uint32 addr, uint32 val) {
  uint32 old = hardware->timer_period;
  switch (addr) {
  case 0:
    logmsg(LOG_MEMORY, 10, "timer: write %x to period\n", val);
    hardware->timer_period = val * 2;
    if (!hardware->timer_active) {
      hardware->timer_active = 1;
      hardware->ev_timer_irq.tick = tick + hardware->timer_period;
      add_event(&(hardware->ev_timer_irq));
    }
    break;
  default:
    logmsg(LOG_MEMORY, 5, "timer: write %x to %x (pc=%8x)\n",
	   val, addr, cpu->reg[15] - 4);
  }
}

static uint32 adc_read(struct hardware *hardware, uint32 adr) {
  switch (adr) {
  case 0: /* most of buttons */
    if (ui_get_key(2)) return 0x280;
    if (ui_get_key(3)) return 0x300;
    if (ui_get_key(4)) return 0x100;
    if (ui_get_key(5)) return 0x20;
    if (ui_get_key(6)) return 0x60;
    if (ui_get_key(7)) return 0x1a0;
    if (ui_get_key(8)) return 0x200;
    return 0x3ff;
  case 4: /* battery */
    return 450;
  case 8:
    if (ui_get_key(1)) return 0;
    return 0x3ff;
  default:
    logmsg(LOG_ADC, 5, "adc: read from %x\n", adr);
  }
  return ~0;
}

static void adc_write(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_ADC, 5, "adc: write %x to %x\n", val, adr);
}

static uint32 codec_read(struct codec *codec, uint32 adr) {
  switch (adr) {
  case 0x398:
    logmsg(LOG_MEMORY, 5, "codec: volume read\n");
    return codec->volume;
  default:
    logmsg(LOG_MEMORY, 5, "codec: read from %x\n", adr);
  }
  return ~0;
}

static void codec_write(struct codec *codec,
			uint32 addr, uint32 val) {
  switch (addr) {
  case 0x398:
    codec->volume = val;
    logmsg(LOG_MEMORY, 5, "codec: volume set to %x\n", val);
    break;
  default:
    logmsg(LOG_MEMORY, 5, "codec: write %x to %x (pc=%8x)\n",
	   val, addr, cpu->reg[15] - 4);
  }
}

static uint32 clocks_read(struct clocks *clocks, uint32 adr) {
  if (adr >= 0 && adr < 0x30) {
    logmsg(LOG_MEMORY, 5, "clocks: read from reg0 %x\n", adr >> 2);
    return clocks->reg0[adr >> 2];
  } else if (adr >= 0x30 && adr < 0x60) {
    logmsg(LOG_MEMORY, 5, "clocks: read from reg30 %x\n", (adr - 0x30) >> 2);
    return clocks->reg30[(adr - 0x30) >> 2];
  } else if (adr >= 0x60 && adr < 0x90) {
    logmsg(LOG_MEMORY, 5, "clocks: read from reg60 %x\n", (adr - 0x60) >> 2);
    return clocks->reg60[(adr - 0x60) >> 2];
  } else if (adr >= 0x90 && adr < 0xc0) {
    logmsg(LOG_MEMORY, 5, "clocks: read from reg90 %x\n", (adr - 0x90) >> 2);
    return clocks->reg90[(adr - 0x90) >> 2];
  } else if (adr >= 0xc0 && adr <= 0xc0 + 72 * 4) {
    logmsg(LOG_MEMORY, 5, "clocks: read from regc0 %x\n", (adr - 0xc0) >> 2);
    return clocks->regc0[(adr - 0xc0) >> 2];
  } else if (adr >= 0x308 && adr <= 0x308 + 66 * 4) {
    logmsg(LOG_MEMORY, 5, "clocks: read from reg308 %x\n", (adr - 0x308) >> 2);
    return clocks->reg308[(adr - 0x308) >> 2];
  } else if (adr >= 0x414 && adr <= 0x414 + 2 * 4) {
    logmsg(LOG_MEMORY, 5, "clocks: read from reg414 %x\n", (adr - 0x414) >> 2);
    return clocks->reg414[(adr - 0x414) >> 2];
  } else if (adr >= 0x420 && adr <= 0x420 + 17 * 4) {
    logmsg(LOG_MEMORY, 5, "clocks: read from reg420 %x\n", (adr - 0x420) >> 2);
    return clocks->reg420[(adr - 0x420) >> 2];
  } else if (adr >= 0xcac && adr <= 0xce0) {
    logmsg(LOG_MEMORY, 5, "clocks: read from regcac %x\n", (adr - 0xcac));
    return ~0;
  } else if (adr >= 0xce4 && adr <= 0xcfc) {
    logmsg(LOG_MEMORY, 5, "clocks: read from regce4 %x\n", (adr - 0xce4));
    return ~0;
  } else {
    logmsg(LOG_MEMORY, 5, "clocks: read from %x\n", adr);
    return -1;
  }
}

static void clocks_write(struct clocks *clocks,
			uint32 adr, uint32 val) {
  if (adr >= 0 && adr < 0x30) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to reg0 %x\n", val, adr >> 2);
    clocks->reg0[adr >> 2] = val;
  } else if (adr >= 0x30 && adr < 0x60) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to reg30 %x\n", val, (adr - 0x30) >> 2);
    clocks->reg30[(adr - 0x30) >> 2] = val;
  } else if (adr >= 0x60 && adr < 0x90) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to reg60 %x\n", val, (adr - 0x60) >> 2);
    clocks->reg60[(adr - 0x60) >> 2] = val;
  } else if (adr >= 0x90 && adr < 0xc0) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to reg90 %x\n", val, (adr - 0x90) >> 2);
    clocks->reg90[(adr - 0x90) >> 2] = val;
  } else if (adr >= 0xc0 && adr <= 0xc0 + 72 * 4) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to regc0 %x\n", val, (adr - 0xc0) >> 2);
    clocks->regc0[(adr - 0xc0) >> 2] = val;
  } else if (adr >= 0x308 && adr <= 0x308 + 66 * 4) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to reg308 %x\n", val, (adr - 0x308) >> 2);
    clocks->reg308[(adr - 0x308) >> 2] = val;
  } else if (adr >= 0x414 && adr <= 0x414 + 2 * 4) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to reg414 %x\n", val, (adr - 0x414) >> 2);
    clocks->reg414[(adr - 0x414) >> 2] = val;
  } else if (adr >= 0x420 && adr <= 0x420 + 17 * 4) {
    int n = (adr - 0x420) >> 2;
    logmsg(LOG_MEMORY, 5, "clocks: write %x (%x %x %x %x) to reg420 %x\n",
	   val,
	   val >> ((n == 0x10) ? 13 : 11),
	   (val >> 3) & ((n == 0x10) ? 0x3ff : 0xff),
	   (val >> 2) & 1,
	   val & 1,
	   n);
    clocks->reg420[(adr - 0x420) >> 2] = val;
  } else if (adr >= 0xcac && adr <= 0xce0) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to regcac %x\n", val, (adr - 0xcac));
  } else if (adr >= 0xce4 && adr <= 0xcfc) {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to regce4 %x\n", val, (adr - 0xce4));
  } else {
    logmsg(LOG_MEMORY, 5, "clocks: write %x to %x\n", val, adr);
  }
}

static uint32 i2c_read(struct hardware *hardware, uint32 adr) {
  switch (adr & 0xff) {
  case 4:
    logmsg(LOG_I2C, 5, "i2c: status read\n", adr);
    return 0;
  default:
    logmsg(LOG_I2C, 5, "i2c: read from %x\n", adr);
  }
  return ~0;
}

static void i2c_write(struct hardware *hardware, uint32 adr, uint32 val) {
  logmsg(LOG_I2C, 5, "i2c: write %x to %x\n", val, adr);
}

static void mmu_set_mapping(struct hardware *hardware,
			    uint32 dest, uint32 src) {
  hardware->mmu[dest >> 21] = src >> 21;
  memory_copy_configuration(hardware->memory, dest, dest + 0x1fffff,
			    hardware->phys_memory, src);
}

static uint32 mmu_read(struct hardware *hardware, uint32 adr) {
  if (adr >= 0x18 && adr < 0x58)
    return hardware->mmu[(adr - 0x18) / 4];
  else {
    logmsg(LOG_MEMORY, 5, "mmu: read from %x\n", adr);
    return 0;
  }
}

static void mmu_write(struct hardware *hardware, uint32 adr, uint32 val) {
  if (adr >= 0x18 && adr < 0x58)
    mmu_set_mapping(hardware, (adr - 0x18) << 19, (val & 0x7ff) << 21);
  else
    logmsg(LOG_MEMORY, 5, "mmu: write %x to register %x\n", val, adr);
}

void initialize_mmu(struct hardware *hardware, struct memory *phys_memory,
		    struct memory *memory) {
  int i;
  memory_copy_configuration(memory, 0, 0xffffffff, phys_memory, 0);
  hardware->memory = memory;
  hardware->phys_memory = phys_memory;
  for (i = 0; i < 16; i++)
    hardware->mmu[i] = i;
  mmu_set_mapping(hardware, 0, 0x10400000);
}

static void dma_irq(struct event *e) {
  struct dma_ctrl *dma = e->data;
  if (!(dma->inten & 1)) {
    dma->intstat |= 1;
    set_irq_request(dma->hardware, 0x1c);
  }
  e->tick = tick + 12345;
  add_event(e);
}

static uint32 dma_read(struct dma_ctrl *dma, uint32 adr) {
  switch (adr) {
  case 0x404:
    logmsg(LOG_MEMORY, 5, "dma: read from intstat\n");
    return dma->intstat;
  case 0x408:
    logmsg(LOG_MEMORY, 5, "dma: read from inten\n");
    return dma->inten;
  default:
    if (adr >= 0 && adr <= 0x100) {
      int ch = adr >> 5;
      logmsg(LOG_MEMORY, 5, "dma(%d): read from %x\n", ch, adr);
      switch (adr & 0x1f) {
      case 0:
	return dma->channel[ch].src;
      case 4:
	return dma->channel[ch].dest;
      case 8:
	return dma->channel[ch].len;
      case 0xc:
	return dma->channel[ch].r0c;
      case 0x10:
	return dma->channel[ch].r10;
      case 0x1c:
	return dma->channel[ch].r1c;
      default:
	return 0;
      }
    } else {
      logmsg(LOG_MEMORY, 5, "dma: read from %x\n", adr);
      return 0;
    }
  }
}

static void dma_write(struct dma_ctrl *dma, uint32 adr, uint32 val) {
  switch (adr) {
  case 0x404:
    logmsg(LOG_MEMORY, 5, "dma: write %x to intstat\n", val);
    dma->intstat = val;
    break;
  case 0x408:
    logmsg(LOG_MEMORY, 5, "dma: write %x to inten\n", val);
    dma->inten = val;
    break;
  default:
    if (adr >= 0 && adr <= 0x100) {
      int ch = adr >> 5;
      logmsg(LOG_MEMORY, 5, "dma(%d): write %x to %x\n", ch, val, adr);
      switch (adr & 0x1f) {
      case 0:
	dma->channel[ch].src = val;
	break;
      case 4:
	dma->channel[ch].dest = val;
	break;
      case 8:
	dma->channel[ch].len = val;
	break;
      case 0xc:
	dma->channel[ch].r0c = val;
	break;
      case 0x10:
	dma->channel[ch].r10 = val;
	break;
      case 0x1c:
	dma->channel[ch].r1c = val;
	break;
      }
    } else
      logmsg(LOG_MEMORY, 5, "dma: write %x to %x\n", val, adr);
  }
}

static void emu_write(struct hardware *hardware, uint32 adr, uint32 val) {
  switch (adr & 0xff) {
  case 0:
    hardware->log_buf[hardware->log_ptr++] = val;
    if (val == 0 || hardware->log_ptr >= sizeof(hardware->log_buf) - 1) {
      hardware->log_buf[sizeof(hardware->log_buf) - 1] = 0;
      logmsg(LOG_GENERAL, 5, "log: %s\n", hardware->log_buf);
      hardware->log_ptr = 0;
    }
    break;

  case 4:
    set_profiling(val);
    break;
  }
}

static uint8 hardware_flash_read(void *data, uint32 adr) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;

  chip = selected_flash(hardware);
  if (chip == -1) {
    logmsg(LOG_FLASH, 5, "flash: no chip selected during read from %x\n", adr);
    return ~0;
  }

  flash = hardware->flash[chip];
  if (!flash)
    return ~0;

  return flash_read(flash, adr & 0xf);
}

static void hardware_flash_write(void *data, uint32 adr, uint8 val) {
  struct hardware *hardware = data;
  struct flash *flash;
  int chip;

  chip = selected_flash(hardware);
  if (chip == -1) {
    logmsg(LOG_FLASH, 5, "flash: no chip selected during write %x to %x\n", val, adr);
    return;
  }

  flash = hardware->flash[chip];
  if (!flash)
    return;

  flash_write(flash, adr & 0xf, val);
}

static uint32 hardware_read_word(void *data, uint32 adr) {
  struct hardware *hardware = data;

  switch ((adr >> 8) & 0xfffc) {
  case 0x30:
    return gpio_read(data, adr);
  case 0x40:
  case 0x44:
  case 0x48:
  case 0x4c:
    return clocks_read(&(hardware->clocks), adr & 0xfff);
  case 0x1040:
    return lcd_read(hardware->lcd, adr);
  case 0x1048:
  case 0x104c:
    return dma_read(&(hardware->dma_ctrl), adr & 0x7ff);
  case 0x1050:
    return mmu_read(hardware, adr & 0xff);
  case 0x3000:
  case 0x3004:
    return read_803(data, adr);
  case 0x24:
    return adc_read(hardware, adr & 0xff);
  case 0x208:
    return i2c_read(hardware, adr & 0xff);
  case 0x2000:
    return codec_read(&(hardware->codec), adr & 0xfff);
  default:
    logmsg(LOG_MEMORY, 10, "read from %8x\n", adr);
    return 0xffffffff;
  }
}

static void hardware_write_word(void *data, uint32 adr, uint32 val) {
  struct hardware *hardware = data;

  switch ((adr >> 8) & 0xfffc) {
  case 0x30:
    gpio_write(data, adr, val);
    break;
  case 0x40:
  case 0x44:
  case 0x48:
  case 0x4c:
    clocks_write(&(hardware->clocks), adr & 0xfff, val);
    break;
  case 0x1040:
    lcd_write(hardware->lcd, adr, val);
    break;
  case 0x1048:
  case 0x104c:
    dma_write(&(hardware->dma_ctrl), adr & 0x7ff, val);
    break;
  case 0x1050:
    mmu_write(hardware, adr & 0xff, val);
    break;
  case 0x3000:
  case 0x3004:
    write_803(data, adr, val);
    break;
  case 0x24:
    adc_write(hardware, adr & 0xff, val);
    break;
  case 0x208:
    i2c_write(hardware, adr & 0xff, val);
    break;
  case 0x200:
    timer_write(hardware, adr & 0xff, val);
    break;
  case 0x2000:
    codec_write(&(hardware->codec), adr & 0xfff, val);
    break;
  case 0x3f00:
    emu_write(hardware, adr & 0xff, val);
    break;
  default:
    logmsg(LOG_MEMORY, 10, "write %8x to %8x (pc=%8x)\n", val, adr, cpu->reg[15] - 4);
  }
}

void install_hardware(struct hardware *hardware,
		      struct memory *memory) {
  set_area(memory, 0x80000000, 0x803fffff,
	   read_void, write_void,
	   read_half_word_void, write_half_word_void,
	   hardware_read_word, hardware_write_word, hardware);
  set_area(memory, 0x20000000, 0x200fffff,
	   prog_flash_read_byte, write_void,
	   prog_flash_read_half_word, prog_flash_write_half_word,
	   prog_flash_read_word, write_word_void,
	   hardware->prog_flash);
  set_area(memory, 0x28000000, 0x2800ffff,
	   hardware_flash_read, hardware_flash_write,
	   read_half_word_void, write_half_word_void,
	   read_word_void, write_word_void, hardware);
  /*
  set_area(memory, 0x24100000, 0x2410ffff,
	   usb_read, usb_write,
	   usb_read_half_word, usb_write_half_word,
	   usb_read_word, usb_write_word, hardware->usb);
  */
}

struct prog_flash *create_prog_flash(char *fname) {
  struct prog_flash *flash;

  flash = xmalloc(sizeof(struct prog_flash));
  if (!flash)
    return NULL;

  flash->state = 0;

  flash->content = xmalloc(1024 * 1024);
  if (!flash->content) {
    free(flash);
    return NULL;
  }
  memset(flash->content, 0xff, 1024 * 1024);

  flash->fp = fopen(fname, "r+");
  if (!flash->fp) {
    error_sys("%s: ", fname);
    free(flash->content);
    free(flash);
    return NULL;
  }

  fread(flash->content, 1, 1024 * 1024, flash->fp);
  return flash;
}

void free_prog_flash(struct prog_flash *flash) {
  if (flash) {
    if (flash->fp)
      fclose(flash->fp);
    if (flash->content)
      free(flash->content);
    free(flash);
  }
}

void codec_reset(struct codec *codec) {
  codec->volume = 0;
}

void clocks_reset(struct clocks *clocks) {
  int i;
  for (i = 0; i < 12; i++) {
    clocks->reg0[i] = 0;
    clocks->reg30[i] = 0;
    clocks->reg60[i] = 0;
    clocks->reg90[i] = 0;
  }
  for (i = 0; i < 73; i++)
    clocks->regc0[i] = 0;
  for (i = 0; i < 67; i++)
    clocks->reg308[i] = 0;
  for (i = 0; i < 3; i++)
    clocks->reg414[i] = 0;
  for (i = 0; i < 18; i++)
    clocks->reg420[i] = 0;
}

void hardware_reset(struct hardware *hardware) {
  int i;
  tick = 0;
  for (i = 0; i < 8; i++) {
    hardware->gpio_out[i] = ~0;
    hardware->gpio_dir[i] = 0;
  }

  hardware->ev_timer_irq.f = timer_irq;
  hardware->ev_timer_irq.data = hardware;
  if (hardware->timer_period) {
    hardware->timer_active = 1;
    hardware->ev_timer_irq.tick = tick + hardware->timer_period;
    add_event(&(hardware->ev_timer_irq));
  }

  hardware->dma_ctrl.hardware = hardware;
  hardware->ev_dma_irq.f = dma_irq;
  hardware->ev_dma_irq.data = &(hardware->dma_ctrl);
  hardware->ev_dma_irq.tick = tick + 12345;
  add_event(&(hardware->ev_dma_irq));

  hardware->irq_request = 0;

  codec_reset(&(hardware->codec));
  clocks_reset(&(hardware->clocks));

  hardware->log_ptr = 0;
}

static int usb_get_connected(void *data, int port, int bit) {
  return ((struct hardware *)data)->usb_connected;
}

static int get_hold(void *data, int port, int bit) {
  return 1;
}

static int get_port6b0(void *data, int port, int bit) {
  return 0;
}

static void set_backlight(void *data, int port, int bit, int v) {
  logmsg(LOG_GPIO, 5, "backlight: set to %d\n", v);
}

static void set_rtc_i2c(void *data, int port, int bit, int v) {
  logmsg(LOG_GPIO, 5, "rtc: i2c %s set to %d\n", bit ? "scl" : "sda", v);
}

static void set_unknown_port(void *data, int port, int bit, int v) {
  logmsg(LOG_GPIO, 5, "gpio: bit %d/%d set to %d\n", port, bit, v);
}

void hardware_connect_flash(struct hardware *hardware,
			    int n, struct flash *flash) {
  hardware->flash[n] = flash;
}


struct hardware *create_hardware(void) {
  int i, j;
  struct hardware *hardware = xmalloc(sizeof(struct hardware));
  if (!hardware) return NULL;

  hardware->prog_flash = create_prog_flash(prog_flash_file);
  if (!hardware->prog_flash)
    goto no_prog_flash;

  for (i = 0; i < 4; i++)
    hardware->flash[i] = NULL;

  for (i = 0; i < 8; i++)
    for (j = 0; j < 32; j++) {
      hardware->gpio_handler[i][j].get = NULL;
      hardware->gpio_handler[i][j].set = NULL;
    }

  gpio_install_handler(hardware, 6, 3,
		       hardware, flash_get_busy, NULL);

  gpio_install_handler(hardware, 7, 0,
		       hardware, usb_get_connected, NULL);

  gpio_install_handler(hardware, 5, 2,
		       hardware, get_hold, NULL);

  gpio_install_handler(hardware, 6, 0,
		       hardware, get_port6b0, set_rtc_i2c);
  gpio_install_handler(hardware, 6, 1,
		       hardware, NULL, set_rtc_i2c);

  gpio_install_handler(hardware, 3, 0,
		       hardware, NULL, set_backlight);

  gpio_install_handler(hardware, 2, 0,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 9,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 19,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 11,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 3, 2,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 8,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 3, 6,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 1, 16,
		       hardware, NULL, set_unknown_port);
  gpio_install_handler(hardware, 5, 0,
		       hardware, NULL, set_unknown_port);

  hardware->lcd = create_lcd();
  if (!hardware->lcd)
    goto no_lcd;


  hardware->usb = create_usb();
  if (!hardware->usb)
    goto no_usb;
  hardware->usb_connected = 0;

  hardware->timer_period = 0;

  return hardware;

 no_usb:
  free_lcd(hardware->lcd);
 no_lcd:
  fclose(hardware->flash[0]->fp);
 no_fp:
  free(hardware->flash[0]);
 no_flash:
  free_prog_flash(hardware->prog_flash);
 no_prog_flash:
  free(hardware);
 err:
  return NULL;

}
