/* iRiver iFP 7XX Emulator
   Copyright (C) 2005 Tomasz Malesinski <tmal@mimuw.edu.pl>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */

#include <curses.h>
#include <stdio.h>

#include "ui.h"
#include "lcd.h"
#include "trainer.h"

#define N_KEYS 10
#define STATUS_LEN 80

static int key_timer[10];
static char status_line[STATUS_LEN + 1];

struct key {
  int term_key;
  int iriver_key;
  int time;
};

static WINDOW *stdscr, *pad;
static pad_x, pad_y;

#define SHORT 10
#define LONG 80

static struct key keys[] = {
  {'q', 0, SHORT},
  {'1', 1, SHORT},
  {'!', 1, LONG},
  {'2', 2, SHORT},
  {'@', 2, LONG},
  {'3', 3, SHORT},
  {'#', 3, LONG},
  {' ', 4, SHORT},
  {'M', 4, LONG},
  {'j', 5, SHORT},
  {'J', 5, LONG},
  {'l', 6, SHORT},
  {'L', 6, LONG},
  {'i', 7, SHORT},
  {'I', 7, LONG},
  {'k', 8, SHORT},
  {'K', 8, LONG},
  {0, 0, 0}
};

static void uicurses_set_status(char *s) {
  strncpy(status_line, s, STATUS_LEN);
  status_line[STATUS_LEN] = 0;
}

static int uicurses_init(void) {
  stdscr = initscr();
  cbreak();
  noecho();
  nonl();
  intrflush(stdscr, FALSE);
  keypad(stdscr, TRUE);
  nodelay(stdscr, 1);

  pad = newpad(LCD_HEIGHT, LCD_WIDTH);
  pad_x = pad_y = 0;

  uicurses_set_status("");

  return 0;
}

static void uicurses_cleanup(void) {
  endwin();
}

static void show_trainer_status(void) {
  int n = trainer_n_possible();
  char tmp[80];

  if (n > 3)
    sprintf(tmp, "trainer: %d locations possible", n);
  else if (n == 0)
    sprintf(tmp, "trainer: no possible locations", n);
  else {
    int p[3];
    int i, j, k;
    trainer_get_possible(p, 3);
    sprintf(tmp, "trainer: possible:%n", &k);
    for (i = 0; i < n; i++) {
      sprintf(tmp + k, " %x%n", p[i] + 0x400000, &j);
      k += j;
    }
  }
  uicurses_set_status(tmp);
}

static void keyboard_handle(void) {
  int ch, i;

  for (i = 0; i < N_KEYS; i++)
    if (key_timer[i]) key_timer[i]--;

  while ((ch = getch()) != ERR) {
    switch (ch) {
    case KEY_DOWN:
      pad_y += 16;
      break;

    case KEY_UP:
      pad_y -= 16;
      if (pad_y < 0) pad_y = 0;
      break;

    case KEY_LEFT:
      pad_x -= 16;
      if (pad_x < 0) pad_x = 0;
      break;

    case KEY_RIGHT:
      pad_x += 16;
      break;

    case 't':
      trainer_start();
      show_trainer_status();
      break;

    case 'c':
      trainer_value_changed();
      show_trainer_status();
      break;

    case 'C':
      trainer_value_not_changed();
      show_trainer_status();
      break;

    default:
      for (i = 0; keys[i].term_key; i++)
	if (ch == keys[i].term_key)
	  key_timer[keys[i].iriver_key] = keys[i].time;
      break;
    }
  }
}

static void uicurses_refresh(struct lcd *lcd) {
  int i, w, h, x, y;

  for (y = 0; y < LCD_HEIGHT - 1; y++)
    for (x = 0; x < LCD_WIDTH; x++)
      mvwaddch(pad, y, x, lcd_get_pixel(lcd, x, y) ? '#' : '.');

  keyboard_handle();
  getmaxyx(stdscr, h, w);

  if (pad_x + w > LCD_WIDTH)
    pad_x = LCD_WIDTH - w;
  if (pad_y + h > LCD_HEIGHT)
    pad_y = LCD_HEIGHT - h;
  
  prefresh(pad, pad_y, pad_x, 0, 0, h - 2, w - 1);
  mvaddstr(h - 1, 0, status_line);
  clrtoeol();
  refresh();
}

static int uicurses_get_key(int key) {
  return key_timer[key] > 0;
}

struct ui uicurses_functions = {
  uicurses_init,
  uicurses_cleanup,
  uicurses_refresh,
  uicurses_get_key,
  uicurses_set_status
};
